{% macro render_field(field) %}
    <div class="form-group">
        {{ field.label }}{{ field(class_='form-control', **kwargs)|safe }}
    </div>
{% endmacro %}

{% macro render_button(field) %}
<div class="form-group">
    {{ field(class_='form-control', **kwargs)|safe }}
</div>
{% endmacro %}

{% macro raise_error(field1, field2) %}
<div class="form-group">
    {% if field1.errors or field2.errors %}
        <p style="color: red">Incorrect Email or Password</p>
    {% endif %}
</div>
{% endmacro %}
